from enum import Enum

INCLUDE_ASSET_PERMISSIONS = False
DELETE_ON_TARGET = False


class ConfigureAssets(Enum):
    # Sites = "sites"
    # Users = "users"
    # Groups = "groups"
    Projects = "projects"
