FIND_WORKBOOK_DATASOURCES = """
{
  "data": {
    "workbooksConnection": {
      "nodes": [
        {
          "name": " ",
          "upstreamDatasources": [
            {
              "name": " ",
              "luid": " "
            }
          ]
        },
        {
          "name": " ",
          "upstreamDatasources": [
            {
              "name": " ",
              "luid": " "
            },
            
          ]
        }
      ]
    }
  }
}
"""
